/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.common.callcontext;

import java.util.UUID;

import org.apache.logging.log4j.CloseableThreadContext;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.stereotype.Component;

import de.ozgcloud.archive.ArchiveManagerConfiguration;
import de.ozgcloud.common.grpc.GrpcUtil;
import io.grpc.ForwardingServerCallListener;
import io.grpc.Metadata;
import io.grpc.ServerCall;
import io.grpc.ServerCall.Listener;
import io.grpc.ServerCallHandler;
import io.grpc.ServerInterceptor;
import lombok.RequiredArgsConstructor;

@Component(ArchiveManagerConfiguration.CALL_CONTEXT_SERVER_INTERCEPTOR_NAME)
@RequiredArgsConstructor
public class CallContextGrpcServerInterceptor implements ServerInterceptor {

	@Qualifier(ArchiveManagerConfiguration.CURRENT_USER_SERVICE_NAME)
	private final CurrentUserService currentUserService;
	@Qualifier(ArchiveManagerConfiguration.CALL_CONTEXT_MAPPER_NAME)
	private final CallContextMapper callContextMapper;

	@Override
	public <A, B> Listener<A> interceptCall(ServerCall<A, B> call, Metadata headers, ServerCallHandler<A, B> next) {
		return new LogContextSettingListener<>(next.startCall(call, headers), headers);
	}

	class LogContextSettingListener<A> extends ForwardingServerCallListener.SimpleForwardingServerCallListener<A> {

		private final String requestId;
		private final Metadata headers;

		public LogContextSettingListener(ServerCall.Listener<A> delegate, Metadata headers) {
			super(delegate);
			this.headers = headers;
			this.requestId = getRequestId(headers);
		}

		String getRequestId(Metadata headers) {
			return GrpcUtil.getRequestId(headers).orElseGet(() -> UUID.randomUUID().toString());
		}

		@Override
		public void onMessage(A message) {
			doSurroundOn(() -> super.onMessage(message));
		}

		@Override
		public void onHalfClose() {
			doSurroundOn(super::onHalfClose);
		}

		@Override
		public void onCancel() {
			doSurroundOn(super::onCancel);
		}

		@Override
		public void onComplete() {
			doSurroundOn(super::onComplete);
		}

		@Override
		public void onReady() {
			doSurroundOn(super::onReady);
		}

		void doSurroundOn(Runnable runnable) {
			SecurityContext securityContext = null;
			try (var ctc = CloseableThreadContext.put(GrpcUtil.KEY_REQUEST_ID, requestId)) {
				securityContext = currentUserService.startAndReturnPreviousSecurityContext(callContextMapper.toCallContextUser(headers, requestId));
				runnable.run();
			} finally {
				currentUserService.resetSecurityContext(securityContext);
			}
		}

	}

}
