/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.historie;

import java.util.function.Function;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Service;

import de.ozgcloud.archive.common.DateConverter;
import de.ozgcloud.archive.vorgang.VorgangWithEingang;
import de.xoev.xdomea.HistorienProtokollInformationType;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
@Service
public class ExportHistorieService {

	private static final String ASSIGN_USER_ORDER = "ASSIGN_USER";
	private final VorgangChangeHistoryService vorgangChangeHistoryService;

	public Stream<HistorienProtokollInformationType> createHistorienProtokollInformationTypes(VorgangWithEingang vorgang) {
		var history = vorgangChangeHistoryService.createVorgangChangeHistory(vorgang);
		return Stream.of(
				history.getStatusChangeHistory().stream().map(this::createHistorienProtokollInformationType),
				history.getAktenzeichenChangeHistory().stream().map(this::createHistorienProtokollInformationType),
				history.getAssignedUserChangeHistory().stream().map(this::createHistorienProtokollInformationType))
				.flatMap(Function.identity());
	}

	HistorienProtokollInformationType createHistorienProtokollInformationType(VorgangChange vorgangChange) {
		var historienProtokollInformationType = new HistorienProtokollInformationType();
		historienProtokollInformationType.setMetadatumAlterWert(createValueBeforeChange(vorgangChange));
		historienProtokollInformationType.setMetadatumNeuerWert(createValueAfterChange(vorgangChange));
		historienProtokollInformationType.setAkteur(createAkteur(vorgangChange));
		historienProtokollInformationType.setDatumUhrzeit(DateConverter.toXmlGregorianCalendar(vorgangChange.getFinishedAt()));
		historienProtokollInformationType.setAktion(vorgangChange.getOrder());
		return historienProtokollInformationType;
	}

	String createAkteur(VorgangChange vorgangChange) {
		return appendOrganisationseinheitenID(vorgangChange.getAuthorFullName(), vorgangChange.getOrganisationseinheitenID());
	}

	String createValueBeforeChange(VorgangChange vorgangChange) {
		return createValueChange(vorgangChange, vorgangChange.getValueBeforeChange());
	}

	String createValueAfterChange(VorgangChange vorgangChange) {
		return createValueChange(vorgangChange, vorgangChange.getValueAfterChange());
	}

	private String createValueChange(VorgangChange vorgangChange, String valueChange) {
		if (isAssignUserOrder(vorgangChange)) {
			return appendOrganisationseinheitenID(valueChange, vorgangChange.getOrganisationseinheitenID());
		}
		return valueChange;
	}

	private boolean isAssignUserOrder(VorgangChange vorgangChange) {
		return vorgangChange.getOrder().equals(ASSIGN_USER_ORDER);
	}

	String appendOrganisationseinheitenID(String text, String organisationseinheitenID) {
		return StringUtils.isNotBlank(text) ? text + "; " + organisationseinheitenID : StringUtils.EMPTY;
	}
}
