/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.vorgang;

import java.util.Optional;
import java.util.function.Predicate;

import javax.xml.datatype.XMLGregorianCalendar;

import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Component;

import de.ozgcloud.archive.common.DateConverter;
import de.xoev.xdomea.GeburtType;
import de.xoev.xdomea.KontaktType;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
@Component
class KontaktTypeCreator {

	private static final Predicate<Antragsteller> HAS_VORNAME = antragsteller -> StringUtils.isNotBlank(antragsteller.getVorname());
	private static final Predicate<Antragsteller> HAS_NACHNAME = antragsteller -> StringUtils.isNotBlank(antragsteller.getNachname());
	private static final Predicate<Antragsteller> HAS_ANREDE = antragsteller -> StringUtils.isNotBlank(antragsteller.getAnrede());
	private static final Predicate<Antragsteller> HAS_GEBURTSDATUM = antragsteller -> geburtsdatumToISO(antragsteller).isPresent();

	private final NameNatuerlichePersonTypeCreator nameNatuerlichePersonTypeCreator;

	public Optional<KontaktType> create(VorgangWithEingang vorgang) {
		return getAntragstellerIfHasRequiredData(vorgang).map(this::toKontaktType);
	}

	Optional<Antragsteller> getAntragstellerIfHasRequiredData(VorgangWithEingang vorgang) {
		return Optional.ofNullable(vorgang.getEingang().getAntragsteller())
				.filter(HAS_VORNAME.or(HAS_NACHNAME).or(HAS_ANREDE).or(HAS_GEBURTSDATUM));
	}

	KontaktType toKontaktType(Antragsteller antragsteller) {
		var kontakt = new KontaktType();
		kontakt.setName(nameNatuerlichePersonTypeCreator.create(antragsteller));

		geburtsdatumToISO(antragsteller)
				.map(this::createGeburtType)
				.ifPresent(kontakt::setGeburt);

		return kontakt;
	}

	static Optional<XMLGregorianCalendar> geburtsdatumToISO(Antragsteller antragsteller) {
		return DateConverter.convertGermanFormatToISO(antragsteller.getGeburtsdatum());
	}

	GeburtType createGeburtType(XMLGregorianCalendar geburtsdatumIso) {
		var geburtType = new GeburtType();
		geburtType.setDatum(geburtsdatumIso);
		return geburtType;
	}

}
