/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.vorgang;

import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.OzgCloudCreateSubCommandsRequest;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.apilib.user.OzgCloudUserId;
import de.ozgcloud.archive.ArchiveManagerConfiguration;
import de.ozgcloud.archive.common.callcontext.CallContextUser;
import de.ozgcloud.archive.common.callcontext.CurrentUserService;
import de.ozgcloud.command.Command;
import de.xoev.xdomea.AkteType;
import de.xoev.xdomea.AllgemeineMetadatenType;
import de.xoev.xdomea.IdentifikationObjektType;
import de.xoev.xdomea.NkAbgabeType;
import de.xoev.xdomea.VorgangType;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
@Service(ArchiveManagerConfiguration.VORGANG_SERVICE_NAME)
public class VorgangService {

	static final String LOCK_VORGANG_ORDER = "LOCK_VORGANG";
	static final String KEY_LOCK_VORGANG_REASON = "reason";
	static final String LOCK_VORGANG_REASON = "Vorgang wird archiviert.";
	static final String SUB_COMMAND_EXECUTION_MODE = "PARALLEL";

	@Qualifier(ArchiveManagerConfiguration.VORGANG_REMOTE_MANAGER_SERVICE_NAME)
	private final VorgangRemoteService vorgangRemoteService;
	@Qualifier(ArchiveManagerConfiguration.OZGCLOUD_COMMAND_SERVICE_NAME)
	private final OzgCloudCommandService ozgCloudCommandService;
	@Qualifier(ArchiveManagerConfiguration.CURRENT_USER_SERVICE_NAME)
	private final CurrentUserService currentUserService;
	private final CommandMapper commandMapper;

	private final VorgangTypeCreator vorgangTypeCreator;
	private final KopfCreator kopfCreator;

	public VorgangWithEingang getVorgang(String vorgangId) {
		return vorgangRemoteService.findVorgangWithEingang(vorgangId);
	}

	public VorgangType createVorgangType(VorgangWithEingang vorgangWithEingang) {
		return vorgangTypeCreator.create(vorgangWithEingang);
	}

	public NkAbgabeType createKopf(VorgangWithEingang vorgangWithEingang) {
		return kopfCreator.createKopf(vorgangWithEingang);
	}

	public AkteType createAkteType(VorgangWithEingang vorgangWithEingang) {
		var akteType = new AkteType();
		akteType.setIdentifikation(createIdentifikationObjektType());
		akteType.setAllgemeineMetadaten(createAllgemeineMetadatenType(vorgangWithEingang));
		return akteType;
	}

	IdentifikationObjektType createIdentifikationObjektType() {
		var identifikationObjektType = new IdentifikationObjektType();
		identifikationObjektType.setID(UUID.randomUUID().toString());
		return identifikationObjektType;
	}

	AllgemeineMetadatenType createAllgemeineMetadatenType(VorgangWithEingang vorgangWithEingang) {
		var allgemeineMetadatenType = new AllgemeineMetadatenType();
		allgemeineMetadatenType.setKennzeichen(Optional.ofNullable(vorgangWithEingang.getAktenzeichen()).orElse(StringUtils.EMPTY));
		return allgemeineMetadatenType;
	}

	public void lockVorgang(Command parentCommand) {
		ozgCloudCommandService.addSubCommands(buildCreateSubCommandsRequest(parentCommand.getId(), buildLockVorgangSubCommand(parentCommand)));
	}

	OzgCloudCommand buildLockVorgangSubCommand(Command command) {
		var commandBuilder = OzgCloudCommand.builder()
				.order(LOCK_VORGANG_ORDER)
				.vorgangId(commandMapper.toOzgCloudVorgangId(command.getVorgangId()))
				.relationId(commandMapper.mapRelationId(command.getVorgangId()))
				.relationVersion(command.getRelationVersion())
				.bodyObject(buildObjectBody());
		getUserId().ifPresent(commandBuilder::createdBy);
		return commandBuilder.build();
	}

	Map<String, Object> buildObjectBody() {
		return Map.of(KEY_LOCK_VORGANG_REASON, LOCK_VORGANG_REASON);
	}

	Optional<OzgCloudUserId> getUserId() {
		return currentUserService.findUser().flatMap(CallContextUser::getUserId).map(commandMapper::toOzgCloudUserId);
	}

	OzgCloudCreateSubCommandsRequest buildCreateSubCommandsRequest(String parentCommandId, OzgCloudCommand command) {
		return OzgCloudCreateSubCommandsRequest.builder()
				.parentId(parentCommandId)
				.subCommand(command)
				.completedIfSubsCompleted(false)
				.executionMode(SUB_COMMAND_EXECUTION_MODE)
				.build();
	}
}
