package de.ozgcloud.archive.archivierung;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.util.UUID;

import jakarta.activation.DataHandler;
import jakarta.activation.FileDataSource;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedConstruction;
import org.mockito.Spy;

import de.ozgcloud.apilib.client_attribute.OzgCloudAccessPermission;
import de.ozgcloud.apilib.client_attribute.OzgCloudClientAttributeService;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangIdMapper;
import de.ozgcloud.archive.common.callcontext.CallContextUser;
import de.ozgcloud.archive.common.xta.XtaService;
import de.ozgcloud.archive.export.ExportService;
import de.ozgcloud.archive.vorgang.VorgangWithEingangTestFactory;
import de.ozgcloud.archive.vorgang.XdomeaXtaFileTestFactory;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.xta.client.model.XtaFile;
import lombok.SneakyThrows;

class ArchiveServiceTest {

	@Spy
	@InjectMocks
	private ArchiveService service;

	@Mock
	private ExportService exportService;
	@Mock
	private XtaService xtaService;
	@Mock
	private OzgCloudClientAttributeService ozgCloudAttributesService;
	@Mock
	private OzgCloudVorgangIdMapper vorgangIdMapper;

	@Nested
	class TestSetVorgangArchiving {

		@Test
		void shouldBuildBooleanClientAttribute() {
			setVorgangArchiving();

			verify(service).buildBooleanClientAttribute(true);
		}

		@Test
		void shouldCallVorgangIdMapper() {
			setVorgangArchiving();

			verify(vorgangIdMapper).fromString(VorgangWithEingangTestFactory.ID);
		}

		@Test
		void shouldSetClientAttribute() {
			var attribute = OzgCloudBooleanClientAttributeTestFactory.create();
			doReturn(attribute).when(service).buildBooleanClientAttribute(true);
			when(vorgangIdMapper.fromString(VorgangWithEingangTestFactory.ID)).thenReturn(OzgCloudVorgangId.from(VorgangWithEingangTestFactory.ID));

			setVorgangArchiving();

			verify(ozgCloudAttributesService).setClientAttribute(attribute, OzgCloudVorgangId.from(VorgangWithEingangTestFactory.ID));
		}

		private void setVorgangArchiving() {
			service.setVorgangArchiving(VorgangWithEingangTestFactory.ID);
		}
	}

	@Nested
	class TestBuildBooleanClientAttribute {

		@Test
		void shouldSetClientName() {
			var attribute = service.buildBooleanClientAttribute(false);

			assertThat(attribute.getClientName()).isEqualTo(CallContextUser.ARCHIVE_MANAGER_CLIENT_NAME);
		}

		@Test
		void shouldSetAccess() {
			var attribute = service.buildBooleanClientAttribute(false);

			assertThat(attribute.getAccess()).isEqualTo(OzgCloudAccessPermission.READ_ONLY);
		}

		@Test
		void shouldSetAttributeName() {
			var attribute = service.buildBooleanClientAttribute(false);

			assertThat(attribute.getAttributeName()).isEqualTo(ArchiveService.ATTRIBUTE_NAME_ANTRAG_ARCHIVING);
		}

		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldSetValue(boolean isArchiving) {
			var attribute = service.buildBooleanClientAttribute(isArchiving);

			assertThat(attribute.getValue()).isEqualTo(isArchiving);
		}
	}

	@Nested
	class TestArchiveVorgang {

		private final Command command = CommandTestFactory.create();
		private final XtaFile xdomeaFile = XdomeaXtaFileTestFactory.create();

		@BeforeEach
		void mock() {
			doReturn(xdomeaFile).when(service).buildXdomeaXtaFile(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallBuildXdomeaXtaFile() {
			archiveVorgang();

			verify(service).buildXdomeaXtaFile(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallSendXdomeaFile() {
			archiveVorgang();

			verify(xtaService).sendXdomeaFile(xdomeaFile);
		}

		private void archiveVorgang() {
			service.archiveVorgang(command);
		}
	}

	@Nested
	class TestBuildXdomeaXtaFile {

		private final String fileNameId = UUID.randomUUID().toString();

		@BeforeEach
		void mock() {
			doReturn(fileNameId).when(service).createFileNameId();
			when(exportService.buildXdomeaFileName(any())).thenReturn(XdomeaXtaFileTestFactory.FILE_NAME);
			doReturn(XdomeaXtaFileTestFactory.DATA_HANDLER).when(service).createFileContent(VorgangWithEingangTestFactory.ID, fileNameId);
		}

		@Test
		void shouldCallCreateFileNameId() {
			buildXdomeaXtaFile();

			verify(service).createFileNameId();
		}

		@Test
		void shouldCallBuildXdomeaFileName() {
			buildXdomeaXtaFile();

			verify(exportService).buildXdomeaFileName(fileNameId);
		}

		@Test
		void shouldCallCreateFileContent() {
			buildXdomeaXtaFile();

			verify(service).createFileContent(VorgangWithEingangTestFactory.ID, fileNameId);
		}

		@Test
		void shouldReturnXdomeaFile() {
			var xdomeaFile = buildXdomeaXtaFile();

			assertThat(xdomeaFile).usingRecursiveComparison().isEqualTo(XdomeaXtaFileTestFactory.create());
		}

		private XtaFile buildXdomeaXtaFile() {
			return service.buildXdomeaXtaFile(VorgangWithEingangTestFactory.ID);
		}
	}

	@Nested
	class TestCreateFileNameId {

		private final UUID uuid = UUID.randomUUID();

		@Test
		void shouldReturnRandomUUID() {
			try (var mockedStaticUUID = mockStatic(UUID.class)) {
				mockedStaticUUID.when(UUID::randomUUID).thenReturn(uuid);

				var returnedFileNameId = service.createFileNameId();

				assertThat(returnedFileNameId).isEqualTo(uuid.toString());
			}
		}
	}

	@Nested
	class TestCreateFileContent {

		private final String fileNameId = UUID.randomUUID().toString();

		@Nested
		class TestOnNoExceptionThrown {
			private MockedConstruction<FileDataSource> fileDataSourceConstruction;
			private FileDataSource fileDataSource;
			private File dataSourceFile;

			private MockedConstruction<DataHandler> dataHandlerConstruction;
			private DataHandler dataHandler;
			private FileDataSource dataHandlerSource;

			@Mock
			private File tempFile;
			@Mock
			private OutputStream outputStream;

			@BeforeEach
			@SneakyThrows
			void setUpMock() {
				doReturn(tempFile).when(service).createTempFile(fileNameId);
				fileDataSourceConstruction = mockConstruction(FileDataSource.class,
						(fileDataSource, context) -> {
							dataSourceFile = (File) context.arguments().get(0);
							this.fileDataSource = fileDataSource;
							when(fileDataSource.getOutputStream()).thenReturn(outputStream);
						});
				dataHandlerConstruction = mockConstruction(DataHandler.class, (dataHandler, context) -> {
					this.dataHandler = dataHandler;
					dataHandlerSource = (FileDataSource) context.arguments().get(0);
				});
			}

			@AfterEach
			void cleanUp() {
				fileDataSourceConstruction.close();
				dataHandlerConstruction.close();
			}

			@Test
			@SneakyThrows
			void shouldCallCreateTempFile() {
				createFileContent();

				verify(service).createTempFile(fileNameId);
			}

			@Test
			void shouldCreateFileDataSourceWithTempFile() {
				createFileContent();

				assertThat(dataSourceFile).isEqualTo(tempFile);
			}

			@Test
			@SneakyThrows
			void shouldGetOutputStreamOfFileDataSource() {
				createFileContent();

				verify(fileDataSource).getOutputStream();
			}

			@Test
			void shouldWriteXdomeaContentToOutputStream() {
				createFileContent();

				verify(exportService).writeXdomeaFileContent(VorgangWithEingangTestFactory.ID, fileNameId, outputStream);
			}

			@Test
			void shouldConstructDataHandlerFromFileDataSource() {
				createFileContent();

				assertThat(dataHandlerConstruction.constructed()).hasSize(1);
				assertThat(dataHandlerSource).isEqualTo(fileDataSource);
			}

			@Test
			void shouldReturnDataHandler() {
				var resultDataHandler = createFileContent();

				assertThat(resultDataHandler).isEqualTo(dataHandler);
			}
		}

		@Nested
		class TestOnIOExceptionThrown {
			@Test
			@SneakyThrows
			void shouldThrowTechnicalExcpetion() {
				doThrow(IOException.class).when(service).createTempFile(any());

				assertThrows(TechnicalException.class, () -> createFileContent());
			}
		}

		private DataHandler createFileContent() {
			return service.createFileContent(VorgangWithEingangTestFactory.ID, fileNameId);
		}
	}

	@Nested
	class TestCreateTempFile {

		private final String fileNameId = UUID.randomUUID().toString();

		@Test
		@SneakyThrows
		void shouldReturnFile() {
			var file = service.createTempFile(fileNameId);

			assertThat(file).isNotNull().isInstanceOf(File.class);
		}
	}

}
