/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.common.callcontext;

import static org.assertj.core.api.AssertionsForClassTypes.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.apache.logging.log4j.ThreadContext;
import org.assertj.core.api.InstanceOfAssertFactories;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;
import org.springframework.security.authentication.AuthenticationTrustResolver;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.common.grpc.GrpcUtil;

class CurrentUserServiceTest {

	@Spy
	@InjectMocks
	private CurrentUserService service;

	@Mock
	private AuthenticationTrustResolver trustResolver;

	@Nested
	class TestGetUser {

		@Test
		void shouldCallFindUser() {
			doReturn(Optional.of(CallContextUserTestFactory.create())).when(service).findUser();

			service.getUser();

			verify(service).findUser();
		}

		@Test
		void shouldReturnUser() {
			var user = CallContextUserTestFactory.create();
			doReturn(Optional.of(user)).when(service).findUser();

			var result = service.getUser();

			assertThat(result).isSameAs(user);
		}

		@Test
		void shouldThrowException() {
			doReturn(Optional.empty()).when(service).findUser();

			assertThatThrownBy(service::getUser).isInstanceOf(IllegalStateException.class);
		}
	}

	@Nested
	class TestFindUser {

		@Mock
		private Authentication auth;

		@BeforeEach
		void init() {
			doReturn(Optional.of(auth)).when(service).findTrustedAuthentication();
		}

		@Test
		void shouldCallFindTrustedAuthentication() {
			service.findUser();

			verify(service).findTrustedAuthentication();
		}

		@Test
		void shouldCallMapToCallContextUser() {
			service.findUser();

			verify(service).mapToCallContextUser(auth);
		}

		@Test
		void shouldReturnUser() {
			var user = CallContextUserTestFactory.create();
			doReturn(user).when(service).mapToCallContextUser(auth);

			var result = service.findUser();

			assertThat(result).contains(user);
		}

		@Test
		void shouldFilterEmptyClientName() {
			doReturn(CallContextUserTestFactory.createBuilder().clientName(null).build()).when(service).mapToCallContextUser(auth);

			var result = service.findUser();

			assertThat(result).isEmpty();
		}
	}

	@Nested
	class TestMapToCallContextUser {

		@Mock
		private Authentication auth;

		@Test
		void shouldReturnUser() {
			var user = CallContextUserTestFactory.create();
			doReturn(user).when(service).buildDefaultCallContextUser(any());

			var result = service.mapToCallContextUser(auth);

			assertThat(result).isSameAs(user);
		}

		@Test
		void shouldCallBuildDefaultCallContextUser() {
			service.mapToCallContextUser(auth);

			verify(service).buildDefaultCallContextUser(auth);
		}

		@Test
		void shouldReturnDefaultUser() {
			var user = CallContextUserTestFactory.create();
			doReturn(user).when(service).buildDefaultCallContextUser(auth);

			var result = service.mapToCallContextUser(auth);

			assertThat(result).isSameAs(user);
		}
	}

	@Nested
	class TestBuildDefaultCallContextUser {

		@Mock
		private Authentication auth;
		@Mock
		private GrantedAuthority grantedAuthority;

		@Test
		void shouldSetClientName() {
			doReturn(CallContextUserTestFactory.CLIENT_NAME).when(auth).getName();

			var result = service.buildDefaultCallContextUser(auth);

			assertThat(result.getClientName()).isEqualTo(CallContextUserTestFactory.CLIENT_NAME);
		}

		@Test
		void shouldSetAccessLimited() {
			when(grantedAuthority.getAuthority()).thenReturn(CallContextUserTestFactory.ORGANISATIONS_EINHEITEN_ID);
			doReturn(List.of(grantedAuthority)).when(auth).getAuthorities();

			var user = service.buildDefaultCallContextUser(auth);

			assertThat(user.getAccessLimitedToOrgaIds()).asInstanceOf(InstanceOfAssertFactories.LIST)
					.containsExactly(CallContextUserTestFactory.ORGANISATIONS_EINHEITEN_ID);
		}

	}

	@Nested
	class TestFindAuthentication {

		@Mock
		private Authentication auth;

		@Test
		void shouldCallFindTrustedAuthentication() {
			service.findAuthentication();

			verify(service).findTrustedAuthentication();
		}

		@Test
		void shouldReturnAuthentication() {
			doReturn(Optional.of(auth)).when(service).findTrustedAuthentication();

			var result = service.findAuthentication();

			assertThat(result).contains(auth);
		}
	}

	@Nested
	class TestFindTrustedAuthentication {

		@Mock
		private SecurityContext context;
		@Mock
		private Authentication auth;

		private MockedStatic<SecurityContextHolder> contextHolderMock;

		@BeforeEach
		void init() {
			contextHolderMock = mockStatic(SecurityContextHolder.class);
			contextHolderMock.when(SecurityContextHolder::getContext).thenReturn(context);
			when(context.getAuthentication()).thenReturn(auth);
		}

		@AfterEach
		void cleanup() {
			contextHolderMock.close();
		}

		@Test
		void shouldCallGetContext() {
			service.findTrustedAuthentication();

			contextHolderMock.verify(SecurityContextHolder::getContext);
		}

		@Test
		void shouldCallGetAuthentication() {
			service.findTrustedAuthentication();

			verify(context).getAuthentication();
		}

		@Test
		void shouldCallIsAnonymous() {
			service.findTrustedAuthentication();

			verify(trustResolver).isAnonymous(auth);
		}

		@Test
		void shouldCallIsAuthenticated() {
			service.findTrustedAuthentication();

			verify(auth).isAuthenticated();
		}

		@Test
		void shouldFilterAnonymous() {
			when(trustResolver.isAnonymous(any())).thenReturn(true);

			var result = service.findTrustedAuthentication();

			assertThat(result).isEmpty();
			verify(auth, never()).isAuthenticated();
		}

		@Test
		void shouldFilterNotAuthenticated() {
			doReturn(false).when(trustResolver).isAnonymous(auth);
			doReturn(false).when(auth).isAuthenticated();

			var result = service.findTrustedAuthentication();

			assertThat(result).isEmpty();
		}
	}

	@DisplayName("Create user")
	@Nested
	class TestCreateUser {

		@Test
		void shouldSetUserId() {
			var user = service.createUser(CommandTestFactory.create());

			assertThat(user.getUserId()).contains(CommandTestFactory.CREATED_BY);
		}

		@Test
		void shouldSetUserName() {
			var createdByName = "user-name";

			var user = service.createUser(CommandTestFactory.createBuilder().createdByName(createdByName).build());

			assertThat(user.getUserName()).contains(createdByName);
		}

		@Test
		void shouldSetClientName() {
			var clientName = "client";

			var user = service.createUser(CommandTestFactory.createBuilder().createdByClientName(clientName).build());

			assertThat(user.getClientName()).isEqualTo(clientName);
		}

		@Test
		void shouldSetAccessLimited() {
			var user = service.createUser(CommandTestFactory.create());

			assertThat(user.isAccessLimited()).isFalse();
		}

		@Test
		void shouldCallGetRequestId() {
			service.createUser(CommandTestFactory.create());

			verify(service).getRequestId();
		}

		@Test
		void shouldSetRequestId() {
			var requestId = UUID.randomUUID().toString();
			doReturn(requestId).when(service).getRequestId();

			var user = service.createUser(CommandTestFactory.create());

			assertThat(user.getRequestId()).isEqualTo(requestId);
		}
	}

	@Nested
	class TestGetRequestId {

		@Test
		void shouldReturnRequestId() {
			var requestId = UUID.randomUUID().toString();
			try (var contextMock = mockStatic(ThreadContext.class)) {
				contextMock.when(ThreadContext::getContext).thenReturn(Map.of(GrpcUtil.KEY_REQUEST_ID, requestId));

				var result = service.getRequestId();

				assertThat(result).isEqualTo(requestId);
			}
		}

		@Test
		void shouldReturnNewRequestId() {
			try (var ctc = mockStatic(ThreadContext.class)) {
				ctc.when(ThreadContext::getContext).thenReturn(Collections.emptyMap());

				var result = service.getRequestId();

				assertThat(result).isNotEmpty();
			}
		}
	}

	@Nested
	class TestStartAndReturnPreviousSecurityContext {

		@Nested
		class TestWithCommand {

			private static final Command COMMAND = CommandTestFactory.create();

			@Mock
			private SecurityContext securityContext;

			@Test
			void shouldCallCreateUser() {
				startAndReturnPreviousSecurityContext();

				verify(service).createUser(COMMAND);
			}

			@Test
			void shouldCallStartAndReturnPreviousSecurityContext() {
				var user = CallContextUserTestFactory.create();
				doReturn(user).when(service).createUser(any());

				startAndReturnPreviousSecurityContext();

				verify(service).startAndReturnPreviousSecurityContext(user);
			}

			@Test
			void shouldReturnContext() {
				doReturn(securityContext).when(service).startAndReturnPreviousSecurityContext(any(CallContextUser.class));

				var result = startAndReturnPreviousSecurityContext();

				assertThat(result).isSameAs(securityContext);
			}

			private SecurityContext startAndReturnPreviousSecurityContext() {
				return service.startAndReturnPreviousSecurityContext(COMMAND);
			}
		}

		@Nested
		class TestWithCallContextUser {

			private final CallContextUser newUser = CallContextUserTestFactory.create();
			private final CallContextAuthenticationToken newAuthentication = CallContextAuthenticationToken.authenticate(newUser);

			@Mock
			private SecurityContext previousSecurityContext;

			@BeforeEach
			void init() {
				SecurityContextHolder.setContext(previousSecurityContext);
			}

			@Test
			void shouldReturnPreviousSecurityContext() {
				var returnedContext = startAndReturnPreviousSecurityContext();

				assertThat(returnedContext).isSameAs(previousSecurityContext);
			}

			@Test
			void shouldSetContextWithNewUser() {
				startAndReturnPreviousSecurityContext();

				var currentAuthentication = SecurityContextHolder.getContext().getAuthentication();
				assertThat(currentAuthentication).usingRecursiveComparison().isEqualTo(newAuthentication);
			}

			private SecurityContext startAndReturnPreviousSecurityContext() {
				return service.startAndReturnPreviousSecurityContext(newUser);
			}
		}
	}

	@Nested
	class TestResetSecurityContext {

		@Mock
		private SecurityContext securityContext;

		private MockedStatic<SecurityContextHolder> contextHolderMock;

		@BeforeEach
		void init() {
			contextHolderMock = mockStatic(SecurityContextHolder.class);
		}

		@AfterEach
		void cleanup() {
			contextHolderMock.close();
		}

		@Test
		void shouldClearContext() {
			resetSecurityContext();

			contextHolderMock.verify(SecurityContextHolder::clearContext);
		}

		@Test
		void shouldSetContext() {
			resetSecurityContext();

			contextHolderMock.verify(() -> SecurityContextHolder.setContext(securityContext));
		}

		@Test
		void shouldNotSetContext() {
			service.resetSecurityContext(null);

			contextHolderMock.verify(() -> SecurityContextHolder.setContext(any()), never());
		}

		private void resetSecurityContext() {
			service.resetSecurityContext(securityContext);
		}
	}
}