/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.common.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.OzgCloudCommandTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.archive.vorgang.VorgangWithEingangTestFactory;
import de.ozgcloud.command.CommandStatus;

class CommandServiceTest {

	@Spy
	@InjectMocks
	private CommandService service;
	@Mock
	private CommandRemoteService remoteService;
	@Mock
	private OzgCloudCommandService ozgCloudCommandService;

	@Nested
	class TestFindFinishedCommands {

		private static final String VORGANG_STATUS_FINISHED = "FINISHED";

		@Test
		void shouldCallRemoteService() {
			service.findFinishedCommands(VorgangWithEingangTestFactory.ID);

			verify(remoteService).findCommands(VorgangWithEingangTestFactory.ID, Optional.of(VORGANG_STATUS_FINISHED), Optional.empty());
		}
	}

	@DisplayName("Find pending")
	@Nested
	class TestFindPending {

		private final String order = LoremIpsum.getInstance().getWords(1);

		@Test
		void shouldCallRemoteService() {
			service.findPending(VorgangWithEingangTestFactory.ID, order);

			verify(remoteService).findCommands(VorgangWithEingangTestFactory.ID, Optional.of(CommandStatus.PENDING.name()),
					Optional.of(order));
		}

		@Test
		void shouldReturnValue() {
			var command = ArchiveManagerCommand.builder().build();
			when(remoteService.findCommands(any(), any(), any())).thenReturn(Stream.of(command));

			var commands = service.findPending(VorgangWithEingangTestFactory.ID, order).toList();

			assertThat(commands).containsExactly(command);
		}
	}

	@Nested
	class TestHasPendingCommandsExceptArchive {

		private final String order = LoremIpsum.getInstance().getWords(1);

		@Test
		void shouldCallOzgCloudCommandService() {
			hasPendingCommandsExceptWithOrder();

			verify(ozgCloudCommandService).getPendingCommands(OzgCloudVorgangId.from(VorgangWithEingangTestFactory.ID));
		}

		@Test
		void shouldReturnFalseOnNoPendingCommands() {
			when(ozgCloudCommandService.getPendingCommands(any())).thenReturn(Collections.emptyList());

			var result = hasPendingCommandsExceptWithOrder();

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnTrueOnPendingCommands() {
			when(ozgCloudCommandService.getPendingCommands(any())).thenReturn(List.of(OzgCloudCommandTestFactory.create()));

			var result = hasPendingCommandsExceptWithOrder();

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseOnCommandWithExcpetionOrder() {
			when(ozgCloudCommandService.getPendingCommands(any())).thenReturn(List.of(OzgCloudCommandTestFactory.createBuilder()
					.order(order)
					.build()));

			var result = hasPendingCommandsExceptWithOrder();

			assertThat(result).isFalse();
		}

		private boolean hasPendingCommandsExceptWithOrder() {
			return service.hasPendingCommandsExceptWithOrder(VorgangWithEingangTestFactory.ID, order);
		}
	}
}
