/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.archive.kommentar;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.archive.common.TimeMapper;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;

class KommentarMapperTest {

	@Spy
	@InjectMocks
	private KommentarMapper mapper = Mappers.getMapper(KommentarMapper.class);
	@Mock
	private GrpcObjectMapper grpcObjectMapper;
	@Mock
	private TimeMapper timeMapper;
	@Mock
	private AttachmentMapper attachmentMapper;

	@Nested
	class TestFromItem {

		@BeforeEach
		void init() {
			doReturn(KommentarTestFactory.KOMMENTAR).when(mapper).mapItemToKommentar(GrpcKommentarTestFactory.ITEM);
		}

		@Test
		void shouldMapItemToKommentar() {
			callMapper();

			verify(mapper).mapItemToKommentar(GrpcKommentarTestFactory.ITEM);
		}

		@Test
		void shouldReturnKommentar() {
			var kommentar = callMapper();

			assertThat(kommentar).isEqualTo(KommentarTestFactory.KOMMENTAR);
		}

		private Kommentar callMapper() {
			return mapper.fromItem(GrpcKommentarTestFactory.create());
		}
	}

	@Nested
	class TestMapItemToKommentar {

		@BeforeEach
		void init() {
			doReturn(GrpcKommentarTestFactory.createItemAsMap()).when(mapper).mapItemToItemMap(GrpcKommentarTestFactory.ITEM);
			doReturn(KommentarTestFactory.KOMMENTAR).when(mapper).mapItemMapToKommentar(GrpcKommentarTestFactory.createItemAsMap());
		}

		@Test
		void shouldMapItemToItemMap() {
			callMapper();

			verify(mapper).mapItemToItemMap(GrpcKommentarTestFactory.ITEM);
		}

		@Test
		void shouldMapItemMapToKommentar() {
			callMapper();

			verify(mapper).mapItemMapToKommentar(GrpcKommentarTestFactory.createItemAsMap());
		}

		@Test
		void shouldReturnKommentar() {
			var kommentar = callMapper();

			assertThat(kommentar).isEqualTo(KommentarTestFactory.KOMMENTAR);
		}

		private Kommentar callMapper() {
			return mapper.mapItemToKommentar(GrpcKommentarTestFactory.ITEM);
		}
	}

	@Nested
	class TestMapItemToItemMap {

		@BeforeEach
		void init() {
			when(grpcObjectMapper.mapFromGrpc(GrpcKommentarTestFactory.ITEM)).thenReturn(GrpcKommentarTestFactory.createItemAsMap());
		}

		@Test
		void shouldMapFromGrpc() {
			callMapper();

			verify(grpcObjectMapper).mapFromGrpc(GrpcKommentarTestFactory.ITEM);
		}

		@Test
		void shouldReturnItemMap() {
			var itemMap = callMapper();

			assertThat(itemMap).isEqualTo(GrpcKommentarTestFactory.createItemAsMap());
		}

		Map<String, Object> callMapper() {
			return mapper.mapItemToItemMap(GrpcKommentarTestFactory.ITEM);
		}
	}

	@Nested
	class TestMapItemMapToKommentar {

		@Nested
		class OnAllFieldsSet {

			@BeforeEach
			void init() {
				when(timeMapper.parseString(any())).thenReturn(KommentarTestFactory.CREATED_AT);
				when(attachmentMapper.mapAttachments(anyList())).thenReturn(KommentarTestFactory.ATTACHMENTS);
			}

			@Test
			void shouldCallTimeMapper() {
				callMapper();

				verify(timeMapper).parseString(KommentarTestFactory.CREATED_AT_STR);
			}

			@Test
			void shouldCallAttachmentMapper() {
				callMapper();

				verify(attachmentMapper).mapAttachments(GrpcKommentarTestFactory.ATTACHMENTS);
			}

			@Test
			void shouldMapAllFields() {
				var kommentar = callMapper();

				assertThat(kommentar).usingRecursiveComparison().isEqualTo(KommentarTestFactory.create());
			}

			private Kommentar callMapper() {
				return mapper.mapItemMapToKommentar(GrpcKommentarTestFactory.createItemAsMap());
			}
		}

		@Nested
		class OnAttachmentsMissing {

			@Test
			void shouldSetToEmptyCollection() {
				var kommentar = mapper.mapItemMapToKommentar(Map.of());

				assertThat(kommentar.getAttachments()).isEmpty();
			}
		}
	}

	@Nested
	class TestMapObjectToString {

		@Mock
		private Object object;

		@Test
		void shouldReturnNullIfObjectIsNull() {
			var result = mapper.mapObjectToString(null);

			assertThat(result).isNull();
		}

		@Test
		void shouldCallToString() {
			mapper.mapObjectToString(object);

			verify(mapper).mapObjectToString(object);
		}

		@Test
		void shouldReturnResultOfToString() {
			when(object.toString()).thenReturn(KommentarTestFactory.TEXT);

			var result = mapper.mapObjectToString(object);

			assertThat(result).isEqualTo(KommentarTestFactory.TEXT);
		}
	}
}
